/*
 * Copyright (C) 2017  Grandstream Networks, Inc.
 * Author: Fanan Chen <fnchen@grandstream.cn>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/uaccess.h>
#include <linux/proc_fs.h>
#include <linux/vmalloc.h>
#include <linux/module.h>
#include <linux/seq_file.h>
#include <linux/crc32.h>
#include <mach/gs_type.h>

#include "flash_rw.h"
#include "mmc_rw.h"

enum otp_value_type {
	OTP_BINARY,
	OTP_STRING,
};

struct otp_entry {
	u16 id;
	u16 type;
	u8 name[32];
	u32 length; /* data length */
	u8 data[];
} **otp_entries;

struct otp_struct {
	u32 checksum;
	u32 magic;
	u32 length;  /* total length */
	u32 num_entries;
	u8 reserved[64];
	u8 data[];
} *otp_ptr;

#define OTP_MAGIC 0x4F545000  /* "OTP\0" */
#define EMMC_RWU_SECTORS (256)

extern struct proc_dir_entry *gxvboard_root;

static struct otp_struct *load_otp_sector(void)
{
	int ret = 0;
	struct otp_struct *otp;
	sector_t start = OTP_SECTOR_OFFSET;
	sector_t total = OTP_SECTOR_SIZE;
	sector_t off;
	sector_t nsecs;

	otp = (struct otp_struct*) vmalloc(OTP_SECTOR_SIZE << 9);
	if (!otp)
		return NULL;

	for (off = 0; off < total;) {
		nsecs = total - off;
		if (nsecs > EMMC_RWU_SECTORS)
			nsecs = EMMC_RWU_SECTORS;
		ret = emmc_read(start + off, nsecs, (char *)otp + (off << 9),
				nsecs << 9);
		if (ret < 0) {
			pr_err("load otp from emmc failed:%d\n", ret);
			break;
		}
		off += nsecs;
	}

	return (ret == 0) ? otp : NULL;
}

static u32 checksum(struct otp_struct *otp)
{
	u32 length;
	unsigned char *p;

	length = otp->length;
	if (length > (OTP_SECTOR_SIZE << 9))
		length = (OTP_SECTOR_SIZE << 9);

	p = ((unsigned char *)otp) + 4; /* skip checksum itself */

	return crc32_le(~0, p, length - 4) ^ ~0;
}

static int otp_proc_show(struct seq_file *m, void *v)
{
	struct otp_entry *entry = (struct otp_entry *)m->private;

	seq_write(m, entry->data, entry->length);

	return 0;
}

static int otp_proc_open(struct inode *inode, struct file *file)
{
	struct otp_entry *entry = (struct otp_entry *)PDE_DATA(inode);

	return single_open(file, otp_proc_show, entry);
}

static const struct file_operations otp_proc_fops = {
	.open		= otp_proc_open,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= single_release,
};

static int create_proc_entries(struct otp_struct *otp)
{
	unsigned char *pos;
	struct proc_dir_entry *dir;
	int i;


	if (otp->magic != OTP_MAGIC) {
		pr_err("No OTP sector found: invalid magic\n");
		return -EINVAL;
	}

	if (otp->length > (OTP_SECTOR_SIZE << 9)) {
		pr_err("Invalid OTP sector: length(%d) > Max(%d)\n",
		       otp->length, OTP_SECTOR_SIZE << 9);
		return -EINVAL;
	}

	if (checksum(otp) != otp->checksum) {
		pr_err("Invalid OTP sector: checksum error!\n");
		return -EINVAL;
	}

	dir = proc_mkdir_data("otp", 0, gxvboard_root, (void *)otp);
	if (!dir)
		return -EFAULT;

	otp_entries = vmalloc(sizeof(struct otp_entry *) * otp->num_entries);
	if (!otp_entries)
		return -ENOMEM;

	pos = (unsigned char *)otp->data;
	for (i = 0; i < otp->num_entries; i++) {
		otp_entries[i] = (struct otp_entry *)pos;
		if (!proc_create_data(otp_entries[i]->name, S_IFREG | S_IRUGO, dir,
				      &otp_proc_fops, (void *)otp_entries[i]))
			return -EFAULT;
		pos += sizeof(struct otp_entry) + otp_entries[i]->length;
	}

	return 0;
}

static int __init otp_init(void)
{
	otp_ptr = load_otp_sector();
	if (!otp_ptr)
		return -ENODEV;

	if (create_proc_entries(otp_ptr) < 0) {
		if (otp_entries)
			vfree(otp_entries);
		vfree(otp_ptr);
	}

	return 0;
}

static void __exit otp_exit(void)
{
	if (otp_ptr) {
		if (otp_entries)
			vfree(otp_entries);
		vfree(otp_ptr);
	}
}

module_init(otp_init);
module_exit(otp_exit);

MODULE_AUTHOR("Fanan Chen <fnchen@grandstream.cn>");
MODULE_DESCRIPTION("Grandstream One Time Programmable driver");
MODULE_LICENSE("GPL");
